/*! 
  \file parameter.cpp
  \brief parameter handling

  \author	Reinhard Enders

  $Revision: 1.7 $
  $Id: parameter.cpp,v 1.7 2006/01/06 12:17:59 reinhard Exp $
  
  Copyright (C) 2004-2006 by Reinhard Enders

  Reinhard.Enders@gmx.de
   
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  http://www.gnu.org/copyleft/gpl.html
   
  \include parameter.cpp
*/

#include <sstream>
#include <fstream>
#include <getopt.h>

#include "parameter.h"

using namespace std;

ParameterMap::ParameterMap()
{
//	cerr << "creating ParameterMap" << endl;
}

// the parameter map is destroyed too early
ParameterMap::~ParameterMap()
{
//	cerr << "deleting parameter map" << endl;
/*
  for ( ParameterMap::iterator mi = begin();
  mi != end(); ++mi ) 
  delete mi->second;
*/
}


ParameterBase::ParameterBase(const string name)
    : mName(name)
{
//	cerr << "creating parameter " << name << endl;

    ParameterMap::iterator mi = getParameterMap().find(name);
    if ( mi != getParameterMap().end() ) 
    {
	cerr << "WARNING parameter '" << name 
	     << "' exists already" << endl;
    }
//	assert(mi == getParameterMap().end());

    getParameterMap()[name] = this;
}

ParameterBase::ParameterBase(const string name, 
			     const string description)
    :mName(name)
    ,mDescription(description)
{
//	cerr << "creating parameter " << name << endl;

    ParameterMap::iterator mi = getParameterMap().find(name);
    if ( mi != getParameterMap().end() ) 
    {
	cerr << "WARNING parameter '" << name 
	     << "' exists already" << endl;
    }
//	assert(mi == getParameterMap().end());

    getParameterMap()[name] = this;
}

// we cannot delete the parameter from the map
// because the map is destroyed before
ParameterBase::~ParameterBase()
{
//	cerr << "deleting parameter: " << mName << endl;
    // doesn't work currently!!
/*
  ParameterMap::iterator mi = getParameterMap().find(mName);
  if ( mi != getParameterMap().end() && mi->second == this ) 
  {
  getParameterMap().erase(mi);
  }
*/
}

ParameterMap& ParameterBase::getParameterMap()
{
    static ParameterMap parameterMap;

    return parameterMap;
}


void ParameterBase::readEnvironment(const string& prefix)
{
    for ( ParameterMap::iterator mi = getParameterMap().begin();
	  mi != getParameterMap().end(); ++mi ) 
    {
	string name = prefix + mi->first;
	char* val = getenv(name.c_str());
	if ( val == 0 ) 
	    continue;

	istringstream valuestream(val);

	cerr << "value: " << val << endl;

	valuestream >> *(mi->second);
    }
}

void ParameterBase::writeEnvironment(ostream& out, bool getopt)
{
    for ( ParameterMap::iterator mi = getParameterMap().begin();
	  mi != getParameterMap().end(); ++mi ) 
    {
	if ( getopt ) 
	{
	    out << "--" << mi->first << "=\"" 
		<< *mi->second << "\"" << endl;
	}
	else 
	{
	    out << mi->first << " \t= " << *mi->second << endl;
	}
	if ( mi->second->mDescription != "" ) 
	    out << " \t" << mi->second->mDescription << endl;

    }
}


void ParameterBase::readResourceFile(const string& resourcefile)
{
    std::ifstream in(resourcefile.c_str());
    if ( !in ) 
    {
	return;
    }
    while ( !in.eof() ) 
    {
	string name;
	in >> name;
	ParameterMap::iterator mi = getParameterMap().find(name);
	if ( mi != getParameterMap().end() ) 
	{
	    in.get();
	    in >> *(mi->second);
	}
    }
}

void ParameterBase::writeResourceFile(const string& resourcefile)
{
    std::ofstream out(resourcefile.c_str());

    for ( ParameterMap::iterator mi = getParameterMap().begin();
	  mi != getParameterMap().end(); ++mi ) 
    {
	out << mi->first << " " << *mi->second << endl;
    }
}


void ParameterBase::setValue(const string& name, const string& val)
{
    ParameterMap::iterator mi = getParameterMap().find(name);
    if ( mi != getParameterMap().end() ) 
    {
	istringstream valuestream(val);
	
	valuestream >> *(mi->second);
    }
    else 
    {
	cerr << "cannot find parameter " << name << endl;
    }

}

void ParameterBase::genLongopts(vector<option>& longopts, int val)
{
    for ( ParameterMap::iterator mi = getParameterMap().begin();
	  mi != getParameterMap().end(); ++mi ) 
    {
	option longopt;
	longopt.name = mi->first.c_str();
	longopt.has_arg = 1;
	longopt.flag = NULL;
	longopt.val = 1;
	longopts.push_back(longopt);
    }
}

template<>
void Parameter<bool>::read(istream& in)
{
    std::string value;
    in >> value;
    if ( value == "true" ) 
	mValue = true;
    else if ( value == "false" ) 
	mValue = false;
    else
	std::cerr << "invalid parameter value: " << value << std::endl;
}

template<>
void Parameter<bool>::write(ostream& out) const
{
    if ( mValue ) 
	out << "true";
    else
	out << "false";
}

template<>
void Parameter<string>::read(istream& in)
{
    int c;
    mValue = "";
    while ( (c = in.get()) && c != '\n' && c != EOF ) 
    {
	mValue += c;
    }
}

FileDialogParameter::FileDialogParameter(string pn, string fs, string sf,
					 string cp, string dir)
    : filters(pn + "Filters", fs)
    , sel_filter(pn + "SelFilter", sf)
    , caption(pn + "Caption", cp)
    , dir(pn + "Dir", dir)
{}


ostream& operator<<(ostream& out, const ParameterBase& p)
{
    p.write(out);
    return out;
}

istream& operator>>(istream& in, ParameterBase& p)
{
    p.read(in);
    return in;
}
