/*! 
  \file pesmgr.cpp
  \brief

  \author R. Enders

  $Revision: 1.25 $
  $Id: pesmgr.cpp,v 1.25 2006/01/06 12:18:00 reinhard Exp $
  
  Copyright (C) 2004-2006 by Reinhard Enders
  Reinhard.Enders@gmx.de

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  http://www.gnu.org/copyleft/gpl.html

  \include pesmgr.cpp
*/


//#define GOPDIT_TRACING

#include "config.h"

#include <sys/time.h>
#include <time.h>

#include "assert.h"
#include "parameter.h"
#include "pesmgr.h"
#include "mpeg2output.h"
#include "psmgr.h"
#include "tymgr.h"
#include "editlist.h"


using std::map;
using std::string;

Parameter<bool> IFramesOnly("IFrames",false);

inline size_t PesMgr::incBuffer()
{
    offset = psmgr->tello();
    for ( unsigned i = 0; i < BUFFER_PREFIX; ++i ) 
    {
	pbuffer[i] = pbuffer[BUFFER_SIZE+i];
    }
    size = psmgr->read(buffer, BUFFER_SIZE);
    mpeg2_buffer (parser, buffer, buffer + size);
    return size;
}

// allows different modes of decodings
// abstract class

class PesAbstractDecoder
{
public:
    PesAbstractDecoder(PesMgr& p)
	: pesmgr(p)
	, ui(p.getUi())
	, parser_info(*mpeg2_info(&p.getParser()))
    {}
    virtual ~PesAbstractDecoder() {}

    virtual bool buffer() { return true; }
    virtual bool sequence() { return true; }
    virtual bool sequenceRepeated() { return true; }
    virtual bool gop() { return true; }
    virtual bool picture() { return true; }
    virtual bool slice1st() { return true; }
    virtual bool picture2nd() { return true; }
    virtual bool slice() { return true; }
    virtual bool end() { return true; }
    virtual bool invalid() { return true; }
    virtual bool invalidEnd() { return true; }
    virtual bool unexpected() { return false; }
protected:
    void         setCurTime();

    PesMgr&       pesmgr;
    GopditUi&     ui;
    const mpeg2_info_t&    parser_info;
};

void PesAbstractDecoder::setCurTime()
{
// ERROR add some frame info as well!!!
    unsigned cur_time = unsigned(((parser_info.gop->hours * 60 
				   + parser_info.gop->minutes) * 60
				  + parser_info.gop->seconds) * 1000);
    cur_time += unsigned((parser_info.gop->pictures * 1000.0) / 
			 pesmgr.getFrameRate());

    pesmgr.setCurTime(cur_time);
}

class PesDummyDecoder : public PesAbstractDecoder
{
public:
    PesDummyDecoder(PesMgr& pparser)
	: PesAbstractDecoder(pparser)
    {}

    bool gop() { return false; }
};


class PesTracer : public PesAbstractDecoder
{
public:
    PesTracer(PesMgr& pparser)
	: PesAbstractDecoder(pparser)
    {}
	
    bool buffer() { printf("buffer\n"); return true; }
    bool sequence() { printf("sequence\n"); return true; }
    bool sequenceRepeated()
    { 
	printf("sequenceRepeated\n"); 
	return true;
    }
    bool gop() { printf("gop\n"); return true; }
    bool picture() { printf("picture\n"); return true; }
    bool slice1st() { printf("slice1st\n"); return true; }
    bool picture2nd() { printf("picture2nd\n"); return true; }
    bool slice() { printf("slice\n"); return true; }
    bool end() { printf("end\n"); return true; }
    bool invalid() { printf("invalid\n"); return true; }
    bool invalidEnd() { printf("invalidEnd\n"); return true; }
    bool unexpected() { printf("unexpected\n"); return true; }
};



class PesGopFinder : public PesAbstractDecoder
{
public:
    PesGopFinder(PesMgr& pparser)
	: PesAbstractDecoder(pparser)
    {}
    bool gop()
    { 
	setCurTime();
	return false;
    }
};


// plays one frame sequence from gop to gop
class PesOutputPlayer : public PesAbstractDecoder
{
public:
    PesOutputPlayer(PesMgr& pparser, GopditOutput& poutput, 
		    long gops, bool safe, 
		    suseconds_t current=0, suseconds_t delta=0)
	: PesAbstractDecoder(pparser)
	, i_frame_found(safe)
	, output(poutput)
	, parser((pparser.getParser()))
	, output_ptr(poutput.output)
	, ucurrent(current)
	, udelta(delta)
    {}
    bool gop()
    {
	setCurTime();
	return false;
    }
    bool sequence()
    {
	return sequenceRep();
    }
    bool picture();
    bool slice() { return sliceEnd(); }
    bool end() { return sliceEnd(); }

    bool sequenceRep();
    bool invalidEnd() { return sliceEnd(); }
protected:
    bool              i_frame_found;
private:
    GopditOutput&     output;
    mpeg2dec_t&      parser;
    vo_instance_t*   output_ptr;
    bool sliceEnd();
    vo_setup_result_t setup_result;
    suseconds_t       ucurrent;
    suseconds_t       udelta;

};

bool PesOutputPlayer::sequenceRep()
{
    if (!parser_info.sequence)
    {
	fprintf(stderr,"Got STATE_SEQUENCE without a sequence!?\n");
	return false;
    }

    if ( !pesmgr.getSequenceInfoValid() ) 
    {
	pesmgr.setSequenceInfo(*(parser_info.sequence));
#ifdef GOPDIT_TRACING
	fprintf(stderr,"frames per second: %.4f\n", 
		pesmgr.getFrameRate());
#endif // GOPDIT_TRACING
    }

#ifdef GOPDIT_TRACING	
    fprintf(stderr,"sequenceRep\n");
#endif // GOPDIT_TRACING

//    fprintf(stderr,"movie screen window %lu\n",(long)ui.getMovieScreenn());

    /* sequence started, so initialize output window */
    if ( output_ptr->setup(output_ptr,
			   parser_info.sequence->width,
			   parser_info.sequence->height,
			   parser_info.sequence->chroma_width,
			   parser_info.sequence->chroma_height,
			   (Window)ui.getMovieScreen(),
			   &setup_result))
    {
	fprintf(stderr, "display setup failed\n");
	return false;
    }
    /* does the output need to be converted? */
    if ( setup_result.convert
	 && mpeg2_convert(&parser, setup_result.convert, NULL ) )
    {
	fprintf (stderr, "color conversion setup failed\n");
	exit (1);
    }
    /* can the output provide an fbuf area? */
    if ( output_ptr->set_fbuf )
    {
	uint8_t *buf[3];
	void *id;
		
	mpeg2_custom_fbuf(&parser, 1);
	output_ptr->set_fbuf(output_ptr, buf, &id);
	mpeg2_set_buf(&parser, buf, id);
	output_ptr->set_fbuf(output_ptr, buf, &id);
	mpeg2_set_buf(&parser, buf, id);
    }
    else if ( output_ptr->setup_fbuf )
    {
	uint8_t *buf[3];
	void *id;
		
	output_ptr->setup_fbuf(output_ptr, buf, &id);
	mpeg2_set_buf(&parser, buf, id);
	output_ptr->setup_fbuf(output_ptr, buf, &id);
	mpeg2_set_buf(&parser, buf, id);
	output_ptr->setup_fbuf(output_ptr, buf, &id);
	mpeg2_set_buf(&parser, buf, id);
    }
    mpeg2_skip (&parser, (output_ptr->draw == NULL));
    return true;
}


bool PesOutputPlayer::picture()
{
#ifdef GOPDIT_TRACING
    fprintf(stderr,"picture\n");
#endif // GOPDIT_TRACING

    if ( output_ptr->set_fbuf )
    {
	uint8_t *buf[3];
	void *id;
		
	output_ptr->set_fbuf(output_ptr, buf, &id);
	assert(id);
	mpeg2_set_buf(&parser, buf, id);
	assert(id);
    }
    if ( output_ptr->start_fbuf )
    {
	if ( parser_info.current_fbuf )
	{
	    if ( parser_info.current_fbuf->id == 0 ) 
	    {
// seems to happen, when the format changes 16:9 <-> 4:3
#ifdef GOPDIT_TRACING
		fprintf(stderr,"buffer id is ZERO!\n");
		fprintf(stderr,"gop: %llx\n",pesmgr.getCurGop());
#endif // GOPDIT_TRACING
		pesmgr.reinitParser();
	    }
	    else 
	    {
		output_ptr->start_fbuf(output_ptr,
				       parser_info.current_fbuf->buf,
				       parser_info.current_fbuf->id);
	    }
	}
	else 
	{
	    fprintf(stderr,"Got STATE_PICTURE without a current_fbuf!?\n");
	    return false;
	}
    }
    return true;
}


bool PesOutputPlayer::sliceEnd()
{

#ifdef GOPDIT_TRACING
    fprintf(stderr,"sliceEnd\n");
#endif // GOPDIT_TRACING

    bool display_frame = false;
    if ( parser_info.display_picture
	&& ((parser_info.display_picture->flags & PIC_MASK_CODING_TYPE)
	    == PIC_FLAG_CODING_TYPE_I) )

    {
	display_frame  = true;
	i_frame_found = true;
    }
    display_frame = i_frame_found 
	&& (display_frame || !IFramesOnly.getValue());

    /* might free frame buffer */
    if ( display_frame && parser_info.display_fbuf 
	 && parser_info.display_fbuf->id != 0 )
    {
	if ( output_ptr->draw )
	{
	    output_ptr->draw (output_ptr, 
			      parser_info.display_fbuf->buf,
			      parser_info.display_fbuf->id);
	}
    }
    if ( output_ptr->discard && parser_info.discard_fbuf )
    {
	output_ptr->discard (output_ptr,
			     parser_info.discard_fbuf->buf,
			     parser_info.discard_fbuf->id);
    }
	
    if ( udelta > 0 ) 
    {
	struct timeval  tv;
	struct timezone tz;
	struct timespec req;
	ucurrent += udelta;
	gettimeofday(&tv, &tz);
	req.tv_sec = 0;
	req.tv_nsec = ((ucurrent - tv.tv_usec) % 1000000) * 1000;

//	fprintf(stderr, "ucurrent %7d  usleep %7d\n", 
//		ucurrent, req.tv_nsec/1000);

	nanosleep(&req, NULL);
	ucurrent = ucurrent % 1000000;
    }

    return true;
}

class PesOutputPlayerOne : public PesOutputPlayer
{
public:
    PesOutputPlayerOne(PesMgr& pparser, GopditOutput& poutput, bool safe)
	: PesOutputPlayer(pparser,poutput,0,safe)
    {}

    bool slice()
    {
	return PesOutputPlayer::slice() && !i_frame_found;
    }
    bool end()
    {
	return PesOutputPlayer::end() && !i_frame_found;
    }
};


PesMgr::PesMgr(GopditUi& pui, GopditOutput& poutput,
	       const char* pfilename, const char* pmoviename)
    : backoff(4096)
//    , psmgr(pui,pfilename)
    , moviename(pmoviename)
    , sequence_info_valid(false)
    , output(poutput)
    , ui(pui)
{
    offset   = 0;
    size     = 0;
    running  = false;

    string filename = pfilename;

    // choose the file type along the extension (no other check right now)

    size_t pos = filename.rfind(".");
    if ( pos != string::npos ) 
    {
	string ext = filename.substr(pos+1);

	if ( ext == "ty" ) 
	{
	    psmgr = new TyMgr(ui,pfilename);
	}
	else 
	{
	    psmgr = new PsMgr(ui,pfilename);
	}
    }
    else
    {
	psmgr = new PsMgr(ui,pfilename);
    }


    buffer   = pbuffer+BUFFER_PREFIX;
    for ( unsigned i = 0; i < BUFFER_PREFIX; ++i ) 
    {
	pbuffer[BUFFER_SIZE+i] = 0xff;
    }
    parser  = mpeg2_init();
}


PesMgr::~PesMgr()
{
    if ( parser != 0 ) 
    {
	mpeg2_close(parser);
    }
    delete psmgr;
}


bool PesMgr::draw()
{
    // some more tests??

    const mpeg2_info_t& parser_info = *mpeg2_info(parser);

    if ( parser_info.display_fbuf != 0 && parser_info.display_fbuf->id != 0 ) 
    {
	output.output->draw (output.output, 
			     parser_info.display_fbuf->buf,
			     parser_info.display_fbuf->id);
	return true;
    }
    return false;
}


void PesMgr::resize(int width, int height)
{
    output.output->resize(output.output, width, height);
}


void PesMgr::findNextGop()
{
    PesGopFinder m2d(*this);
    abstractDecode(m2d);
    if ( !isValidCurGop() ) 
    {
	abstractDecode(m2d);
    }
    if ( !isValidCurGop() ) 
    {
	fprintf(stderr,"no valid gop package header\n");
    }
}

/*
  After 'play' in normal mode, the current gop is the gop, where
  playing stopped. If we only back up one gop and play the whole
  sequence of frames until we find a gop again, we are exactly,
  where we were before, so we have to back up 2 gops to achieve
  any 'progress'.
*/
void PesMgr::findPrevGop(unsigned n)
{
//    assert( isValidCurGop() );

    off_t prev_gop = getValidCurGop();
    for ( ; n > 0; --n ) 
    {
	prev_gop = getPrevGop(prev_gop);
    }

    PesGopFinder m2d(*this);

    setOffset(prev_gop);
    abstractDecode(m2d);

    assert( isValidCurGop() );
}


ClipPoint PesMgr::getLastClipPoint()
{
    off_t cur_gop = getCurGop();
    ClipPoint cp = ClipPointSearcher(this).getLastClip();

//    std::cout << "CP off: " << cp.off << " ms: " << cp.ms << std::endl;

    setOffset(cur_gop);
    return cp;
}


off_t PesMgr::getNextGop(off_t offset)
{
    // do we have to save the current gop???
    off_t cur_gop = getCurGop();
    off_t next_gop;

// setOffset initializes aleady the buffer and seeks too much
//    setOffset(offset);

    psmgr->seeko(offset,SEEK_SET);

    psmgr->searchGop();
    
//    assert( isValidCurGop() );
    next_gop = getCurGop();
    setOffset(cur_gop);
    return next_gop;
}


unsigned PesMgr::getTime(off_t gop_offset)
{
    // do we have to save the current gop???
    off_t cur_gop = getCurGop();
    unsigned ms_time;
    PesGopFinder m2d(*this);

    setOffset(gop_offset);
    abstractDecode(m2d);

    ms_time = cur_time;
    setOffset(cur_gop);

    return ms_time;
}

/*
  hopefully previous gaps are mostly cached

  Searching gops should be done low level by the psmgr, we don't need
  the mpeg2 parser for it.
*/
off_t PesMgr::getPrevGop(off_t cur_gop)
{
    assert( isValidOffset(cur_gop) );

    off_t prev_gop = cur_gop;

    map<off_t,off_t>::iterator p = prev_cache.find(cur_gop);
    if ( p != prev_cache.end() ) 
    {
	prev_gop = p->second;
    }
    else 
    {
//	PesGopFinder m2d(*this);
	off_t next_gop;
	off_t back_off = backoff/2;

	// search for a gop that precedes the current one
	do 
	{
	    back_off *= 2;
	    prev_gop = cur_gop < back_off ? 0 : cur_gop - back_off;


	    setOffset(prev_gop);

//	    abstractDecode(m2d);
	    psmgr->searchGop();

//	    assert( isValidCurGop() );
	    prev_gop = getValidCurGop();
	} while ( prev_gop >= cur_gop && back_off < cur_gop );
	
	// if we haven't found any previous gop let's give up
	if ( prev_gop == cur_gop ) 
	{
	    return prev_gop;
	}
	
	// find the immediatly preceding gop
	next_gop = prev_gop;
	do 
	{
	    prev_gop = next_gop;

//	    abstractDecode(m2d);
	    psmgr->searchGop();

	    next_gop = getValidCurGop();
	    // second condition is a hack for end of file handling
	} while ( next_gop != cur_gop && next_gop != prev_gop );
	
	// add to cache
	// we could clear the cache now and then
	// but there are usually only tens of thousand gops
	// which shouldn' be a problem for nowadays memory
	// hashing would be faster, but the bottleneck is surely
	// disk io and the mpeg2 parser
/*
	if ( prev_cache > 2000 ) 
	{
	    prev_cache.clear();
	}
*/
	prev_cache[cur_gop] = prev_gop;
	
	// adjust backoff
	assert( prev_gop < cur_gop );
	if ( backoff < 2*(cur_gop - cur_gop) ) 
	{
	    backoff = 2*(cur_gop - cur_gop);
	}
	
    }
    return prev_gop;
}



void PesMgr::play(long gops, bool safe)
{
    // during initialization, we don't have a valid gop

    assert( safe || isValidCurGop() );

    struct timeval  tv;
    struct timezone tz;
    gettimeofday(&tv, &tz);

    off_t prev_gop;
    running = true;
    PesOutputPlayer m2op(*this,output,gops,safe,tv.tv_usec,
			 gops < 0 ? suseconds_t(1000000 / getFrameRate())  : 0);
    for ( ; running && (gops != 0); --gops ) 
    {
	prev_gop = getCurGop();
	abstractDecode(m2op);
	if ( !isValidCurGop() ) 
	{
	    return;
	}
	if ( prev_gop < getCurGop() )
	{
	    prev_cache[getCurGop()] = prev_gop;
	}
	ui.updateCurrentTime(cur_time);
	ui.updateMovieSlider(getCurGop());

    }
    running = false;
}


// play one slice!!!
void PesMgr::playOneSlice(bool safe)
{
    PesOutputPlayerOne m2op(*this,output,safe);
    abstractDecode(m2op);
    if ( !isValidCurGop() ) 
    {
	fprintf(stderr,"no valid gop package header\n");
    }
    ui.updateCurrentTime(cur_time);
}


bool PesMgr::setOffset(off_t off)
{
    for ( unsigned i = 0; i < BUFFER_PREFIX; ++i ) 
    {
	pbuffer[BUFFER_SIZE + i] = 0xff;
    }
    size   = 0;
    psmgr->seeko(off,SEEK_SET);
#ifdef HAVE_MPEG2_RESET
    // don't know, if I need a full reset
    mpeg2_reset(parser,0);
#ifdef GOPDIT_TRACING
    fprintf(stderr,"parser is reset at %llx\n",off);
#endif // GOPDIT_TRACING
#endif
    incBuffer();
    return true;
}

void PesMgr::abstractDecode(PesAbstractDecoder& m2d)
{
    bool cont = true;
    do 
    {
	switch ( mpeg2_parse(parser) ) 
	{
	case STATE_BUFFER:
	    incBuffer();
	    cont = (size > 0 && m2d.buffer());
	    break;
	case STATE_SEQUENCE:
	    cont = m2d.sequence();
	    break;
	case STATE_SEQUENCE_REPEATED:
	    cont = m2d.sequenceRepeated();
	    break;
	case STATE_GOP:
	    cont = m2d.gop();
	    break;
	case STATE_PICTURE:
	    cont = m2d.picture();
	    break;
	case STATE_SLICE_1ST:
	    cont = m2d.slice1st();
	    break;
	case STATE_PICTURE_2ND:
	    cont = m2d.picture2nd();
	    break;
	case STATE_SLICE:
	    cont = m2d.slice();
	    break;
	case STATE_END:
	    cont = m2d.end();
	    break;
	case STATE_INVALID:
	    cont = m2d.invalid();
	    break;
	case STATE_INVALID_END:
	    cont = m2d.invalidEnd();
	    break;
	default:
	    cont = m2d.unexpected();
	}
    } while ( cont );
}

void PesMgr::reinitParser()
{
    if ( parser != 0 ) 
    {
	mpeg2_close(parser);
    }
    parser = mpeg2_init();
}


off_t PesMgr::getCurGop() const
{
    off_t cur_gop = psmgr->getGopPh();
    if ( ! isValidOffset(cur_gop) )
    {
	ui.report(std::string("found invalid gop"));
    }
    else 
    {
	ui.report(std::string(""));
    }
    return cur_gop;
}


off_t PesMgr::getValidCurGop() const
{
    off_t cur_gop = psmgr->getGopPh();
    if ( ! isValidOffset(cur_gop) )
    {
	ui.report(std::string("found invalid gop"));
	cur_gop = getFsize();
    }
    else 
    {
	ui.report(std::string(""));
    }
    return cur_gop;
}





bool  PesMgr::isValidCurGop() const
{
    return isValidOffset(getCurGop());
}

float PesMgr::getFrameRate() const
{
    if ( !getSequenceInfoValid() ) 
    {
	return 30;
    }
     return double(27000000) / double(sequence_info.frame_period);
}

int PesMgr::getBitRate() const
{
    if ( !getSequenceInfoValid() ) 
    {
	return 0;
    }
    return sequence_info.byte_rate * 8;
}

int PesMgr::getPWidth() const
{
    if ( !getSequenceInfoValid() ) 
    {
	return 0;
    }
    return sequence_info.width;
}

int PesMgr::getPHeight() const
{
    if ( !getSequenceInfoValid() ) 
    {
	return 0;
    }
    return sequence_info.height;
}


void PesMgr::dumpInfo()
{
#ifdef GOPDIT_TRACING
    if ( !getSequenceInfoValid() ) 
    {
	return;
    }

    using std::cerr;
    using std::endl;

    cerr << "Sequence Info" << endl;
    cerr << "width " << sequence_info.width << endl;
    cerr << "height " << sequence_info.height << endl;
    cerr << "chroma_width " << sequence_info.chroma_width << endl;
    cerr << "chroma_height " << sequence_info.chroma_height << endl;
    cerr << "picture_width " << sequence_info.picture_width << endl;
    cerr << "picture_height " << sequence_info.picture_height << endl;
    cerr << "display_width " << sequence_info.display_width << endl;
    cerr << "display_height " << sequence_info.display_height << endl;
    cerr << "pixel_width " << sequence_info.pixel_width << endl;
    cerr << "pixeL_height " << sequence_info.pixel_height << endl;
    cerr << "byte_rate " << sequence_info.byte_rate << endl;
    cerr << "vbv_buffer_size " << sequence_info.vbv_buffer_size << endl;
    cerr << "flags " << sequence_info.flags << endl;
    cerr << "frame_period " << sequence_info.frame_period << endl;
    cerr << "profile_level_id " << sequence_info.profile_level_id << endl;
    cerr << "colour_primaries " << sequence_info.colour_primaries << endl;
    cerr << "transfer_characteristics " 
	 << sequence_info.transfer_characteristics << endl;
    cerr << "matrix_coefficients " << sequence_info.matrix_coefficients << endl;
#endif // GOPDIT_TRACING


}
